from abc import ABC, abstractmethod
from urllib.parse import urlparse

import requests

from pytaku.conf import config

from .exceptions import (
    SourceSite5xxError,
    SourceSite404Error,
    SourceSiteTimeoutError,
    SourceSiteUnexpectedError,
)


def create_session():
    session = requests.Session()
    session.headers["User-Agent"] = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.71 Safari/537.36"
    )
    return session


class Site(ABC):
    def __init__(self):
        self._session = create_session()

    @abstractmethod
    def get_title(self, title_id) -> dict:
        pass

    @abstractmethod
    def get_chapter(self, title_id, chapter_id) -> dict:
        pass

    @abstractmethod
    def search_title(self, query) -> list[dict]:
        pass

    @abstractmethod
    def title_cover(self, title_id, cover_ext) -> str:
        pass

    @abstractmethod
    def title_thumbnail(self, title_id, cover_ext) -> str:
        pass

    @abstractmethod
    def title_source_url(self, title_id) -> str:
        pass

    # optional abstract method
    def login(self, username, password):
        raise NotImplementedError()

    def _http_request(self, method, url, *args, **kwargs):
        headers = kwargs.get("headers", {})
        if "timeout" not in kwargs:
            kwargs["timeout"] = 15

        # print(">>", url, args, kwargs)

        # Proxy shit
        if config.OUTGOING_PROXY_NETLOC:
            parsed_url = urlparse(url)
            url = parsed_url._replace(
                netloc=config.OUTGOING_PROXY_NETLOC,
                scheme="https",
            ).geturl()
            headers["X-Proxy-Target-Host"] = parsed_url.netloc
            headers["X-Proxy-Key"] = config.OUTGOING_PROXY_KEY
            headers["X-Proxy-Scheme"] = parsed_url.scheme
            kwargs["headers"] = headers

        request_func = getattr(self._session, method)
        try:
            resp = request_func(url, *args, **kwargs)
        except requests.exceptions.Timeout:
            raise SourceSiteTimeoutError(url)

        if resp.status_code == 403:
            self._session = create_session()

        if 500 <= resp.status_code <= 599:
            raise SourceSite5xxError(url, resp.status_code, resp.text)
        elif resp.status_code == 404:
            raise SourceSite404Error(url, resp.text)
        elif resp.status_code != 200:
            raise SourceSiteUnexpectedError(url, resp.status_code, resp.text)

        return resp

    def http_get(self, *args, **kwargs):
        return self._http_request("get", *args, **kwargs)

    def http_post(self, *args, **kwargs):
        return self._http_request("post", *args, **kwargs)



import html
import re

import bbcode

from mangoapi.base_site import Site

MANGAPLUS_GROUP_ID = 9097
WEB_COMIC_TAG_ID = "e197df38-d0e7-43b5-9b09-2842d0c326dd"

_bbparser = bbcode.Parser()
_bbparser.add_simple_formatter(
    "spoiler", "<details><summary>Spoiler</summary>%(value)s</details>"
)


class Mangadex(Site):
    def get_title(self, title_id):
        md_resp = self.http_get(
            f"https://api.mangadex.org/manga/{title_id}",
            params={"includes[]": "cover_art"},
        )
        assert md_resp.status_code == 200
        md_json = md_resp.json()
        attrs = md_json["data"]["attributes"]

        is_web_comic = False
        for tag in attrs["tags"]:
            if tag["id"] == WEB_COMIC_TAG_ID:
                is_web_comic = True
                break

        cover = None
        for rel in md_json["data"]["relationships"]:
            if rel["type"] == "cover_art":
                cover = rel["attributes"]["fileName"]

        descriptions = attrs["description"]
        if not descriptions:
            description = ""
        elif "en" in descriptions:
            description = descriptions["en"]
        else:
            description = list(descriptions.values())[0]

        title = {
            "id": title_id,
            "name": list(attrs["title"].values())[0],
            "site": "mangadex",
            "cover_ext": cover,
            "alt_names": [list(alt.values())[0] for alt in attrs["altTitles"]],
            "descriptions": [_bbparser.format(html.unescape(description).strip())]
            if description
            else [],
            "descriptions_format": "html",
            "is_webtoon": is_web_comic,
            "chapters": self.get_chapters_list(title_id),
        }
        return title

    def get_chapters_list(self, title_id):
        chapters = []
        offset = 0
        limit = 100  # max allowed by mangadex api; will be 400 if we try any higher
        while True:
            resp = self.http_get(
                "https://api.mangadex.org/chapter",
                params={
                    "manga": title_id,
                    "translatedLanguage[]": "en",
                    "order[chapter]": "desc",
                    "offset": offset,
                    "limit": limit,
                },
            )
            assert resp.status_code == 200
            body = resp.json()
            chapters += [
                {
                    "id": chap["id"],
                    "name": chap["attributes"]["title"],
                    "groups": [],  # TODO
                    "volume": chap["attributes"]["volume"],
                    **_parse_chapter_number(chap["attributes"]["chapter"]),
                }
                for chap in body["data"]
            ]

            offset += limit
            if len(chapters) < limit or offset >= body["total"]:
                break

        return chapters

    def get_chapter(self, title_id, chapter_id):
        md_resp = self.http_get(f"https://api.mangadex.org/chapter/{chapter_id}")
        assert md_resp.status_code == 200
        md_json = md_resp.json()
        data = md_json["data"]

        title_id = None
        for rel in data["relationships"]:
            if rel["type"] == "manga":
                title_id = rel["id"]
                break
        # chapter_hash = data["attributes"]["hash"]
        # filenames = data["attributes"]["data"]
        md_server = "https://uploads.mangadex.org"

        mdah_api_url = f"https://api.mangadex.org/at-home/server/{chapter_id}"
        mdah_resp = self.http_get(mdah_api_url)
        assert mdah_resp.status_code == 200, f"Failed request to {mdah_api_url}"
        mdah_data = mdah_resp.json()
        mdah_server = mdah_data["baseUrl"]
        chapter_hash = mdah_data["chapter"]["hash"]
        filenames = mdah_data["chapter"]["data"]

        chapter = {
            "id": chapter_id,
            "title_id": title_id,
            "site": "mangadex",
            "name": data["attributes"]["title"],
            "pages": [
                f"{md_server}/data/{chapter_hash}/{filename}" for filename in filenames
            ],
            "pages_alt": [
                f"{mdah_server}/data/{chapter_hash}/{filename}"
                for filename in filenames
            ]
            if mdah_server is not None and mdah_server != md_server
            else [],
            "groups": [],  # TODO
            **_parse_chapter_number(data["attributes"]["chapter"]),
        }
        return chapter

    def search_title(self, query):
        params = {
            "limit": 100,
            "title": query,
            "includes[]": "cover_art",
            "order[relevance]": "desc",
        }
        md_resp = self.http_get("https://api.mangadex.org/manga", params=params)
        assert md_resp.status_code == 200
        results = md_resp.json()["data"]

        titles = []
        for result in results:
            data = result
            cover = None
            for rel in result["relationships"]:
                if rel["type"] == "cover_art":
                    cover = rel["attributes"]["fileName"]
            titles.append(
                {
                    "id": data["id"],
                    "name": data["attributes"]["title"]["en"],
                    "site": "mangadex",
                    "thumbnail": f"https://uploads.mangadex.org/covers/{data['id']}/{cover}.256.jpg",
                }
            )

        return titles

    def title_cover(self, title_id, cover_ext):
        return f"https://uploads.mangadex.org/covers/{title_id}/{cover_ext}.256.jpg"

    def title_thumbnail(self, title_id, cover_ext):
        return f"https://uploads.mangadex.org/covers/{title_id}/{cover_ext}.256.jpg"

    def title_source_url(self, title_id):
        return f"https://mangadex.org/title/{title_id}"


# Titles regex slightly adapted from https://github.com/md-y/mangadex-full-api
# Thanks!
TITLES_PATTERN = re.compile(
    r"""<a[^>]*href=["']\/title\/(\d+)\/\S+["'][^>]*manga_title[^>]*>([^<]*)<"""
)


def _parse_chapter_number(string: str):
    if string in (None, "none"):
        # most likely a oneshot
        return {"number": "0.0", "num_major": 0, "num_minor": 0}
    nums = string.split(".", maxsplit=1)
    result = {"number": string}
    result["num_major"] = nums[0]
    if len(nums) == 2:
        result["num_minor"] = nums[1]
    return result



from secrets import token_urlsafe

from goodconf import Field, GoodConf


class Config(GoodConf):
    model_config = {"default_files": ["pytaku.conf.json"]}

    FLASK_SECRET_KEY: str = Field(initial=lambda: token_urlsafe(50))

    MANGA_HOURS_UNTIL_OUTDATED: int = Field(default=6)

    PROXY_PREFIX: str = Field(default="")
    # ^ use this to e.g. point to a CDN's domain
    PROXY_CACHE_DIR: str = Field(default="proxy_cache")
    PROXY_CACHE_MAX_SIZE: int = Field(default=1024 * 1024 * 1024 * 5)  # 5GiB in bytes
    PROXY_CACHE_MAX_AGE: int = Field(default=3600 * 24 * 2)  # 2 days in seconds

    OUTGOING_PROXY_NETLOC: str = Field()
    OUTGOING_PROXY_KEY: str = Field()


config = Config()



import textwrap
from abc import ABC, abstractmethod
from pathlib import Path


class Storage(ABC):
    @abstractmethod
    def save(self, path: Path, blob: bytes):
        pass

    @abstractmethod
    def exists(self, path: Path) -> bool:
        pass

    @abstractmethod
    def read(self, path: Path) -> bytes:
        pass


class FilesystemStorage(Storage):
    def save(self, path: Path, blob: bytes):
        path = self._split_long_filename(path)
        if not path.parent.is_dir():
            path.parent.mkdir(parents=True)
        path.write_bytes(blob)

    def exists(self, path: Path) -> bool:
        path = self._split_long_filename(path)
        return path.is_file()

    def read(self, path: Path) -> bytes:
        path = self._split_long_filename(path)
        return path.read_bytes()

    @staticmethod
    def _split_long_filename(path: Path) -> Path:
        filename = path.name
        if len(filename) <= 255:
            return path
        else:
            parts = textwrap.wrap(filename, width=255)
            newpath = path.parent
            for part in parts:
                newpath /= Path(part)
            return newpath


# TODO: support other storages e.g. s3-like
storage = FilesystemStorage()



import re
from html import unescape

from mangoapi.base_site import Site

regexes = {
    "title_name": re.compile(r"<title>\s*([^|]+) | Weeb Central</title>"),
    "title_chapter_id": re.compile(
        r'a href="https:\/\/weebcentral.com\/chapters\/([A-Z0-9]+)"'
    ),
    "title_chapters": re.compile(
        r'href="https://weebcentral\.com/chapters/([0-9A-Z]+)".*?>(.*?(([0-9]+)(\.([0-9])+)?))</a>'
    ),
    "title_current_chapter": re.compile(
        r'<button id="selected_chapter".*?>(.*?(([0-9]+)(\.([0-9])+)?))</button>'
    ),
    "title_desc": re.compile(r"<strong>Description<\/strong>\w*[^<]+<p[^>]*>([^<]+)<"),
    "chapter_number": re.compile(r"<span>(.*?(([0-9]+)(\.([0-9])+)?))</span>"),
    "chapter_imgs": re.compile(r'<img\s+src="(https?://.*?)"'),
    "search_data": re.compile(
        r'<a href="https://weebcentral.com/series/([A-Z0-9]+)/(\n|.)+?alt="(.+?) cover"'
    ),
}


class Weebcentral(Site):
    search_table = None

    def __init__(self, keyval_store=None):
        super().__init__()
        self.keyval_store = keyval_store

    def get_title(self, title_id):
        resp = self.http_get(f"https://weebcentral.com/series/{title_id}")
        html = resp.text
        title_name = unescape(regexes["title_name"].search(html).group(1)).strip()
        desc = unescape(regexes["title_desc"].search(html).group(1)).strip()

        # Originally I scraped the "Show All Chapters" AJAX endpoint to get all
        # chapters, but that endpoint is too slow and even times out on Martial Peak
        # which has 3500+ chapters, so I switched to the "chapter select" endpoint
        # instead.

        # The chapter-select endpoint requires a current_chapter url param, so let's
        # pick the last one (typically Chapter 1), which is more stable than picking the
        # latest chapter that can change between requests (unlikely but it doesn't hurt
        # to be foolproof).
        chapter_1_id = regexes["title_chapter_id"].findall(html)[-1]
        chapters_html = self.http_get(
            f"https://weebcentral.com/series/{title_id}/chapter-select?current_chapter={chapter_1_id}&current_page=0"
        ).text

        chapters = [
            {
                "id": chap_id,
                "name": "",
                "volume": "",
                "groups": [],
                "num_major": int(num_major),
                "num_minor": int(num_minor) if num_minor else 0,
                "number": number,
            }
            for chap_id, name, number, num_major, _, num_minor in regexes[
                "title_chapters"
            ].findall(chapters_html)
        ]

        chapter_1_data = regexes["title_current_chapter"].search(chapters_html).groups()
        name, number, num_major, _, num_minor = chapter_1_data
        chapters.append(
            {
                "id": chapter_1_id,
                "name": "",
                "volume": "",
                "groups": [],
                "num_major": int(num_major),
                "num_minor": int(num_minor) if num_minor else 0,
                "number": number,
            }
        )

        return {
            "id": title_id,
            "name": title_name,
            "site": "weebcentral",
            "cover_ext": "webp",
            "is_webtoon": False,
            "chapters": chapters,
            "alt_names": [],
            "descriptions": desc.split("\n\n"),
            "descriptions_format": "text",
        }

    def get_chapter(self, title_id, chapter_id):
        html = self.http_get(f"https://weebcentral.com/chapters/{chapter_id}").text
        name, number, num_major, _, num_minor = (
            regexes["chapter_number"].search(html).groups()
        )

        imgs_html = self.http_get(
            f"https://weebcentral.com/chapters/{chapter_id}/images?is_prev=False&current_page=1&reading_style=long_strip"
        ).text
        pages = regexes["chapter_imgs"].findall(imgs_html)

        result = {
            "id": chapter_id,
            "title_id": title_id,
            "site": "weebcentral",
            "name": "",
            "pages": pages,
            "pages_alt": [],
            "groups": [],
            "num_major": int(num_major),
            "num_minor": int(num_minor) if num_minor else 0,
            "number": number,
        }
        return result

    def search_title(self, query):
        html = self.http_post(
            "https://weebcentral.com/search/simple?location=main",
            data={"text": query},
        ).text

        return [
            {
                "id": title_id,
                "name": unescape(name),
                "site": "weebcentral",
                "thumbnail": self.title_thumbnail(title_id, "webp"),
            }
            for title_id, _, name in regexes["search_data"].findall(html)
        ]

    def title_cover(self, title_id, cover_ext):
        return self.title_thumbnail(title_id, cover_ext)

    def title_thumbnail(self, title_id, cover_ext):
        return f"https://temp.compsci88.com/cover/normal/{title_id}.webp"

    def title_source_url(self, title_id):
        return f"https://weebcentral.com/series/{title_id}"



import time
import traceback
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from json.decoder import JSONDecodeError
from pathlib import Path

from mangoapi.exceptions import SourceSite5xxError, SourceSite404Error
from requests.exceptions import ReadTimeout

from .conf import config
from .persistence import delete_expired_tokens, find_outdated_titles, save_title
from .source_sites import get_title

now = datetime.now


def main_loop():
    workers = [UpdateOutdatedTitles(), DeleteExpiredTokens(), PruneProxyCache()]

    while True:
        for worker in workers:
            if worker.should_run():
                print("Running", worker.__class__.__name__)
                try:
                    worker.run()
                    worker.after_run()
                except Exception:
                    stacktrace = traceback.format_exc()
                    print(stacktrace)
                    worker.after_error(stacktrace)
        time.sleep(5)


class Worker(ABC):
    interval = timedelta(days=1)

    def __init__(self):
        self.last_run = datetime(1, 1, 1)
        self.error_count = 0

    def should_run(self):
        return now() - self.last_run >= self.interval

    @abstractmethod
    def run(self):
        pass

    def after_run(self):
        self.last_run = now()

    def after_error(self, stacktrace):
        # TODO: email or send stacktrace to an ops chat channel or something
        self.error_count += 1

        # If failed repeatedly: give up this run
        if self.error_count > 3:
            self.last_run = now()
            self.error_count = 0


class UpdateOutdatedTitles(Worker):
    interval = timedelta(hours=2)

    def run(self):
        outdated_titles = find_outdated_titles()
        print(f"Found {len(outdated_titles)} outdated titles")
        for title in outdated_titles:
            print(
                f"Updating title {title['id']} - {title['name'][:36]} from {title['site']}...",
                end="",
            )
            try:
                updated_title = get_title(title["site"], title["id"])
                save_title(updated_title)
                chapters = updated_title["chapters"]
                print(f" done ({chapters and chapters[0]['number']})")
                if title["site"] == "mangasee":
                    time.sleep(2)
            except (
                # TODO: on 404, maybe delete title that's no longer on source site?
                SourceSite404Error,
                SourceSite5xxError,
                ReadTimeout,
                JSONDecodeError,
            ) as e:
                print(" skipped because of server error:", e.__class__.__name__, str(e))


class DeleteExpiredTokens(Worker):
    interval = timedelta(days=1)

    def run(self):
        num_deleted = delete_expired_tokens()
        if num_deleted > 0:
            print("Deleted", num_deleted, "tokens")


class PruneProxyCache(Worker):
    """
    If proxy cache dir size exceeds config.PROXY_CACHE_MAX_SIZE,
    delete files that are older than config.PROXY_CACHE_MAX_AGE.

    Only applies for FilesystemStorage.
    TODO: update this accordingly when a new Storage class is introduced.
    """

    interval = timedelta(days=1)

    def run(self):
        cache_dir = Path(config.PROXY_CACHE_DIR)
        cache_size = get_dir_size(cache_dir)

        if cache_size <= config.PROXY_CACHE_MAX_SIZE:
            return

        now = time.time()
        files_deleted = 0
        bytes_deleted = 0
        for child in cache_dir.iterdir():
            if child.is_file():
                stat = child.stat()
                modified_at = stat.st_mtime
                if (now - modified_at) > config.PROXY_CACHE_MAX_AGE:
                    child.unlink()  # yes this means delete
                    files_deleted += 1
                    bytes_deleted += stat.st_size

        if files_deleted > 0:
            in_mb = bytes_deleted / 1024 / 1024
            print(f"Deleted {files_deleted} files ({in_mb:.2f} MiB).")
        else:
            print("Deleted nothing.")


def get_dir_size(path: Path):
    """
    In bytes.
    """
    return sum(f.stat().st_size for f in path.glob("**/*") if f.is_file())



import apsw

DBNAME = "db.sqlite3"

_conn = None


def _row_trace(cursor, row):
    """
    Customize each result row's representation:
    - If query only asks for 1 field, return that result directly instead of tuple
    - If more than 1 field, return dict instead of tuple
    - Because SQLite stores booleans as 0 or 1, convert them to proper python bools
    """
    desc = cursor.getdescription()
    if len(desc) == 1:
        return row[0]
    else:
        result = {}
        for i, (col_name, col_type) in enumerate(desc):
            value = row[i]
            if col_type == "boolean":
                value = bool(value)
            result[col_name] = value
        return result


def get_conn():
    global _conn
    if not _conn:
        _conn = apsw.Connection(DBNAME)
        # Apparently you need to enable this pragma per connection:
        _conn.cursor().execute("PRAGMA foreign_keys = ON;")
        # No idea what the default db busy timeout is, but apparently it's super strict:
        # got BusyError almost consistently when clicking "finish" on latest chapter,
        # making FE send both a "read" and "get title" request at roughly the same time.
        # It still makes no sense though, since "get title" is supposedly a read-only
        # operation and only "read" is a write. According to docs, WAL mode allows 1
        # writer and unlimited readers at the same time. WTF guys?
        # But anyway, until I can get to the bottom of it, let's slap on a band-aid:
        _conn.setbusytimeout(1000)
        _conn.setrowtrace(_row_trace)
    return _conn


def run_sql(*args, return_num_affected=False, **kwargs):
    cursor = run_sql_on_demand(*args, **kwargs)
    if return_num_affected:
        return cursor.execute("select changes();").fetchone()
    return list(cursor)


def run_sql_on_demand(*args, **kwargs):
    return get_conn().cursor().execute(*args, **kwargs)


def run_sql_many(*args, **kwargs):
    return get_conn().cursor().executemany(*args, **kwargs)



import datetime
import subprocess
from importlib import resources
from pathlib import Path

from . import migrations
from .common import DBNAME, get_conn, run_sql

"""
Forward-only DB migration scheme held together by duct tape.

- Uses `user_version` pragma to figure out what migrations are pending.
- Migrations files are in the form `./migrations/mXXXX.sql`.
"""


def _get_current_version():
    return run_sql("PRAGMA user_version;")[0]


def _get_version(migration: Path):
    return int(migration.name[len("m") : -len(".sql")])


def _get_pending_migrations(migrations_dir: Path):
    current_version = _get_current_version()
    migrations = sorted(migrations_dir.glob("m*.sql"))
    return [
        migration
        for migration in migrations
        if _get_version(migration) > current_version
    ]


def _read_migrations(paths):
    """Returns list of (version, sql_text) tuples"""
    results = []
    for path in paths:
        with open(path, "r") as sql_file:
            results.append((_get_version(path), sql_file.read()))
    return results


def _write_db_schema_script(migrations_dir: Path):
    schema = subprocess.run(
        ["sqlite3", DBNAME, ".schema"], capture_output=True, check=True
    ).stdout
    with open(migrations_dir / Path("latest_schema.sql"), "wb") as f:
        f.write(b"-- This file is auto-generated by the migration script\n")
        f.write(b"-- for reference purposes only. DO NOT EDIT.\n\n")
        f.write(schema)


def migrate(overwrite_latest_schema=True):
    # If there's no existing db, create one with the correct pragmas
    if not Path(DBNAME).is_file():
        run_sql("PRAGMA journal_mode = WAL;")

    with resources.path(migrations, "__init__.py") as migrations_dir:
        migrations_dir = migrations_dir.parent
        pending_migrations = _get_pending_migrations(migrations_dir)
        if not pending_migrations:
            print("Nothing to migrate.")
            exit()
        print(f"There are {len(pending_migrations)} pending migrations.")
        migration_contents = _read_migrations(pending_migrations)

        conn = get_conn()
        cursor = conn.cursor()

        # Backup first
        current_version = _get_current_version()
        if current_version != 0:
            now = datetime.datetime.utcnow().isoformat("T", "milliseconds")
            backup_filename = f"db_backup_v{current_version}_{now}.sqlite3"
            print(f"Backup up to {backup_filename}...", end="")
            cursor.execute("VACUUM main INTO ?;", (backup_filename,))
            print(" done")

        # Start migrations
        # NOTE: this is NOT done in a transaction.
        # You'll need to do transactions inside your sql scripts.
        # This is to allow for drastic changes that require temporarily turning off the
        # foreign_keys pragma, which doesn't work inside transactions.
        # If anything goes wrong here, let it abort the whole script. You can always
        # restore from the backup file.
        cursor = conn.cursor()
        for version, sql in migration_contents:
            print("Migrating version", version, "...")
            cursor.execute(sql)
            cursor.execute(f"PRAGMA user_version = {version};")

        if overwrite_latest_schema:
            _write_db_schema_script(migrations_dir)

        print("All done. Current version:", _get_current_version())



import traceback
from concurrent.futures import ThreadPoolExecutor, as_completed

from mangoapi import get_site_class

from .persistence import KeyvalStore

"""
This module adapts mangoapi's API to a more convenient one for app-wide use.
States are all handled here, exposing only a functional API to the rest of the app.
"""

_site_objs = {}


def _get_site(name):
    global _site_objs
    site = _site_objs.get(name)
    if not site:
        site_class = get_site_class(name)
        assert site_class is not None
        site = site_class()
        if name == "mangasee":
            site.keyval_store = KeyvalStore
    return site


def get_chapter(site_name, title_id, chapter_id):
    return _get_site(site_name).get_chapter(title_id, chapter_id)


def get_title(site_name, title_id):
    return _get_site(site_name).get_title(title_id)


def search_title(site_name, query):
    return _get_site(site_name).search_title(query)


def title_cover(site_name, title_id, cover_ext):
    return _get_site(site_name).title_cover(title_id, cover_ext)


def title_thumbnail(site_name, title_id, cover_ext):
    return _get_site(site_name).title_thumbnail(title_id, cover_ext)


def title_source_url(site_name, title_id):
    return _get_site(site_name).title_source_url(title_id)


def search_title_all_sites(query):
    """
    Returns dict in the form of {site_name: List[Title]}
    I should really look into proper type annotations huh.
    """
    site_names = ("weebcentral", "mangadex")
    results = []

    def safe_search(site_name, query):
        try:
            return search_title(site_name, query)
        except Exception:
            print(f"{site_name}'s search function shat the bed for query [{query}]:")
            traceback.print_exc()
            return []

    # Concurrently search from multiple sites.
    # Obviously network I/O bound so a thread pool is appropriate enough.
    with ThreadPoolExecutor(max_workers=5) as executor:
        future_to_site_name = {
            executor.submit(safe_search, site_name, query): site_name
            for site_name in site_names
        }
        for future in as_completed(future_to_site_name):
            site_results = future.result()
            site_name = future_to_site_name[future]
            results.append({"site": site_name, "titles": site_results})

    return results



import subprocess

from mangoapi.mangasee import Mangasee
from pytaku.database.common import get_conn, run_sql
from pytaku.persistence import save_title

ms = Mangasee()


def fetch_title(title_id: str):
    title = ms.get_title(title_id)
    updates = [(title_id, ch["number"], ch["id"]) for ch in title["chapters"]]
    return title, updates


def migrate():
    mangasee_titles = run_sql(
        "SELECT id FROM title WHERE site = 'mangasee' ORDER BY lower(id);"
    )
    print(f"There are {len(mangasee_titles)} titles to update.")

    diffs = []
    new_titles = []
    for title_id in mangasee_titles:
        print(f">> Fetching {title_id}")
        new_title, new_title_diffs = fetch_title(title_id)
        diffs += new_title_diffs
        new_titles.append(new_title)

    print("Diffs:")
    for diff in diffs:
        print(diff)

    print("Starting db transaction")
    conn = get_conn()
    cursor = conn.cursor()
    cursor.execute("pragma foreign_keys = off;")
    cursor.execute("begin transaction;")

    for new_title in new_titles:
        print(f'Saving title {new_title["id"]}')
        save_title(new_title)

    for title_id, old_chapter_id, new_chapter_id in diffs:
        print("Updating", title_id, old_chapter_id, "to", new_chapter_id)
        cursor.execute(
            "UPDATE chapter SET id=? WHERE id=? AND title_id=? AND site='mangasee';",
            (new_chapter_id, old_chapter_id, title_id),
        )
        cursor.execute(
            "UPDATE read SET chapter_id=? WHERE chapter_id=? AND title_id=? AND site='mangasee';",
            (new_chapter_id, old_chapter_id, title_id),
        )

    cursor.execute("pragma foreign_key_check;")
    cursor.execute("commit;")
    cursor.execute("pragma foreign_keys = on;")
    print("All done!")


def main():
    subprocess.run(["systemctl", "--user", "stop", "pytaku"], check=True)
    subprocess.run(["systemctl", "--user", "stop", "pytaku-scheduler"], check=True)

    migrate()

    subprocess.run(["systemctl", "--user", "start", "pytaku"], check=True)
    subprocess.run(["systemctl", "--user", "start", "pytaku-scheduler"], check=True)


if __name__ == "__main__":
    main()
